# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Tools for parsing neon model definition files (YAML formatted) and
generating neon model objects from the definition.
"""
from copy import deepcopy
import numpy as np
import yaml

from neon import NervanaObject
from neon.layers import GeneralizedCost
from neon.models import Model
import neon.optimizers
from neon.layers.container import Sequential


def format_yaml_dict(yamldict, type_prefix):
    """
    Helper function for format the YAML model config into
    the proper format for object and layer initialization

    Arguments:
        yamldict (dict): dictionary with model parameters

        type_prefix (str): module path for this object

    Returns:
        dict : formatted dict
    """
    yamldict['type'] = type_prefix + yamldict['type']
    return yamldict


def create_objects(root_yaml,
                   be_type='gpu',
                   batch_size=128,
                   rng_seed=None,
                   device_id=0,
                   default_dtype=np.float32,
                   stochastic_rounding=False):
    """
    Instantiate objects as per the given specifications.

    Arguments:
        root_yaml (dict): Model definition dictionary parse from YAML file

        be_type (str): backend either 'gpu', 'mgpu' or 'cpu'

        batch_size (int): Batch size.
        rng_seed (None or int): random number generator seed

        device_id (int): for GPU backends id of device to use

        default_dtype (type): numpy data format for default data types,

        stochastic_rounding (bool or int): number of bits for stochastic rounding
                                           use False for no rounding

    Returns:
        tuple: Contains model, cost and optimizer objects.
    """

    assert NervanaObject.be is not None, 'Must generate a backend before running this function'

    # can give filename or parse dictionary
    if type(root_yaml) is str:
        with open(root_yaml, 'r') as fid:
            root_yaml = yaml.safe_load(fid.read())

    # in case references were used
    root_yaml = deepcopy(root_yaml)

    # initialize layers
    yaml_layers = root_yaml['layers']

    # currently only support sequential in yaml
    layer_dict = {'layers': yaml_layers}
    layers = Sequential.gen_class(layer_dict)

    # initialize model
    model = Model(layers=layers)

    # cost (before layers for shortcut derivs)
    cost_name = root_yaml['cost']
    cost = GeneralizedCost.gen_class({'costfunc': {'type': cost_name}})

    # create optimizer
    opt = None
    if 'optimizer' in root_yaml:
        yaml_opt = root_yaml['optimizer']
        typ = yaml_opt['type']
        opt = getattr(neon.optimizers, typ).gen_class(yaml_opt['config'])

    return model, cost, opt



# ----------------------------------------------------------------------------
# Copyright 2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import sys
import numpy as np
from cffi import FFI
from neon import NervanaObject
from neon.transforms import Rectlinclip

ffi = FFI()
omp_threshold = 100000


def isccontiguous(x):
    return x.flags['C_CONTIGUOUS'] is True


def isfcontiguous(x):
    return x.flags['F_CONTIGUOUS'] is True


def iscontiguous(x):
    return isccontiguous(x) or isfcontiguous(x)


def isscalar(x):
    return isinstance(x, np.float) or isinstance(x, np.int) \
           or isinstance(x, np.float32)


def issame_buffer(x1, x2):
    return x1.__array_interface__['data'][0] == x2.__array_interface__['data'][0]


def isfloat_tensor(x):
    return isinstance(x, np.ndarray) and x.dtype == np.float32


def istensor1d(x):
    return (x.ndim == 1) or (x.ndim == 2 and x.shape[1] == 1)


def istensor2d(x):
    return x.ndim == 2 and x.shape[1] != 1


def blas_copy(x1, x2):
    if issame_buffer(x1, x2):
        x1[:] = x2
    elif isfloat_tensor(x1) and isfloat_tensor(x2) \
            and isccontiguous(x1) and isccontiguous(x2) and x1.shape == x2.shape \
            and x1.size > omp_threshold:
        _x1 = ffi.cast("float *", ffi.from_buffer(x1))
        _x2 = ffi.cast("float *", ffi.from_buffer(x2))
        NervanaObject.be.mathlib.cmath_copy(_x2, x1.size, _x1)
    else:
        x1[:] = x2


def neg(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_neg(_x, x.size, _y)
        return y
    else:
        return -x


def sqrt(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_sqrt(_x, x.size, _y)
        return y
    else:
        return np.sqrt(x)


def square(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_square(_x, x.size, _y)
        return y
    else:
        return np.square(x)


def exp(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_exp(_x, x.size, _y)
        return y
    else:
        return np.exp(x)


def log(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_log(_x, x.size, _y)
        return y
    else:
        return np.log(x)


def safelog(x):
    if isfloat_tensor(x) and isccontiguous(x) \
            and x.size > omp_threshold:
        y = np.empty_like(x)
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_safelog(_x, x.size, _y)
        return y
    else:
        return np.log(np.maximum(x, np.exp(-50.)))


def add(x1, x2):
    if isfloat_tensor(x1) and isfloat_tensor(x2) \
            and isccontiguous(x1) and isccontiguous(x2):
        if x1.shape == x2.shape and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_add(_x1, _x2, x1.size, _y)
            return y
        elif istensor2d(x1) and istensor1d(x2) \
                and x1.shape[0] == x2.shape[0] and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_addmv(_x1, _x2, x1.shape[0], x1.shape[1], _y)
            return y
        elif istensor1d(x1) and istensor2d(x2) \
                and x1.shape[0] == x2.shape[0] and x2.size > omp_threshold:
            y = np.empty_like(x2)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_addmv(_x2, _x1, x2.shape[0], x2.shape[1], _y)
            return y
        else:
            return np.add(x1, x2)
    else:
        return np.add(x1, x2)


def sub(x1, x2):
    if isfloat_tensor(x1) and isfloat_tensor(x2) \
            and isccontiguous(x1) and isccontiguous(x2):
        if x1.shape == x2.shape and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_sub(_x1, _x2, x1.size, _y)
            return y
        elif istensor2d(x1) and istensor1d(x2) \
                and x1.shape[0] and x2.shape[0] and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_submv(_x1, _x2, x1.shape[0], x1.shape[1], _y)
            return y
        else:
            return np.subtract(x1, x2)
    else:
        return np.subtract(x1, x2)


def mul(x1, x2):
    if isfloat_tensor(x1) and isscalar(x2) \
            and isccontiguous(x1) and x1.size > omp_threshold:
        y = np.empty_like(x1)
        _x1 = ffi.cast("float *", ffi.from_buffer(x1))
        _x2 = ffi.cast("double", x2)
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_mul(_x1, x1.size, _x2, _y)
        return y
    elif isscalar(x1) and isfloat_tensor(x2) \
            and isccontiguous(x2) and x2.size > omp_threshold:
        y = np.empty_like(x2)
        _x1 = ffi.cast("double", x1)
        _x2 = ffi.cast("float *", ffi.from_buffer(x2))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_mul(_x2, x2.size, _x1, _y)
        return y
    elif isfloat_tensor(x1) and isfloat_tensor(x2)\
            and isccontiguous(x1) and isccontiguous(x2):
        if x1.shape == x2.shape and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_mulmm(_x1, _x2, x1.size, _y)
            return y
        elif istensor2d(x1) and istensor1d(x2)\
                and x1.shape[0] == x2.shape[0] and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_mulmv(_x1, _x2, x1.shape[0], x1.shape[1], _y)
            return y
        elif istensor1d(x1) and istensor2d(x2)\
                and x1.shape[0] == x2.shape[0] and x2.size > omp_threshold:
            y = np.empty_like(x2)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_mulmv(_x2, _x1, x2.shape[0], x2.shape[1], _y)
            return y
        else:
            return np.multiply(x1, x2)
    else:
        return np.multiply(x1, x2)


def div(x1, x2):
    if isfloat_tensor(x1) and isscalar(x2) \
            and isccontiguous(x1) and x1.size > omp_threshold:
        y = np.empty_like(x1)
        _x1 = ffi.cast("float *", ffi.from_buffer(x1))
        _x2 = ffi.cast("double", x2)
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_div(_x1, x1.size, _x2, _y)
        return y
    elif isfloat_tensor(x1) and isfloat_tensor(x2)\
            and isccontiguous(x1) and isccontiguous(x2):
        if x1.shape == x2.shape and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_divmm(_x1, _x2, x1.size, _y)
            return y
        elif istensor2d(x1) and istensor1d(x2)\
                and x1.shape[0] == x2.shape[0] and x1.size > omp_threshold:
            y = np.empty_like(x1)
            _x1 = ffi.cast("float *", ffi.from_buffer(x1))
            _x2 = ffi.cast("float *", ffi.from_buffer(x2))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_divmv(_x1, _x2, x1.shape[0], x1.shape[1], _y)
            return y
        else:
            return np.divide(x1, x2)
    else:
        return np.divide(x1, x2)


def _force_corder(x):
    if x.flags.f_contiguous:
        return (x.T, 't')
    else:
        return (x, 'n')


def blas_dot(a, b, c=None, alpha=1.0, beta=0):
    # convert non-contiguous tensor into contiguous to speed up
    if not iscontiguous(a):
        a = np.ascontiguousarray(a, dtype=np.float32)
    if not iscontiguous(b):
        b = np.ascontiguousarray(b, dtype=np.float32)
    if (iscontiguous(a) and iscontiguous(b) and c is None) \
            or (iscontiguous(a) and iscontiguous(b) and iscontiguous(c)) \
            and sys.version_info < (3, 4):
        m = a.shape[0]
        n = b.shape[1]
        k = a.shape[1]
        a_corder, transa = _force_corder(a)
        b_corder, transb = _force_corder(b)

        if c is None:
            c = np.zeros((m, n), dtype=np.float32)

        lda = k if transa is 'n' else m
        ldb = n if transb is 'n' else k
        ldc = n

        # TODO: cffi cast not compatible with f_contiguous array with py3.4
        _a = ffi.cast("float *", ffi.from_buffer(a))
        _b = ffi.cast("float *", ffi.from_buffer(b))
        _c = ffi.cast("float *", ffi.from_buffer(c))
        _transa = ffi.cast("char", transa)
        _transb = ffi.cast("char", transb)
        NervanaObject.be.mathlib.cmath_gemm(
            _transa, _transb, m, n, k, alpha, _a, lda, _b, ldb, beta, _c, ldc)
    else:
        if beta == 0:
            np.dot(a, b, c)
            if alpha != 1:
                np.multiply(c, alpha, c)
            return c
        if beta != 1:
            np.multiply(c, beta, c)
        tmp = np.empty(c.shape, dtype=c.dtype)
        np.dot(a, b, tmp)
        if alpha != 1:
            np.multiply(tmp, alpha, tmp)
        np.add(c, tmp, c)
    return c


def sum(x, axis=None, keepdims=True):
    if isfloat_tensor(x) and isccontiguous(x) and istensor2d(x) \
            and x.size > omp_threshold and axis is not None and keepdims is True:
        if axis == 1:
            y = np.zeros((x.shape[0], 1), dtype=x.dtype)
            _x = ffi.cast("float *", ffi.from_buffer(x))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_sum(_x, 1, x.shape[0], x.shape[1], _y)
            return y
        elif axis == 0:
            y = np.zeros((1, x.shape[1]), dtype=x.dtype)
            _x = ffi.cast("float *", ffi.from_buffer(x))
            _y = ffi.cast("float *", ffi.from_buffer(y))
            NervanaObject.be.mathlib.cmath_sum(_x, 0, x.shape[0], x.shape[1], _y)
            return y
        else:
            raise AttributeError('axis of a 2d tensor should only be 0 or 1')
    else:
        return np.sum(x, axis=axis, keepdims=keepdims)


def blas_axpby(a, x, b, y):
    if isfloat_tensor(x) and isfloat_tensor(y) and isccontiguous(x) \
       and isccontiguous(y) and x.shape == y.shape:
        _x = ffi.cast("float *", ffi.from_buffer(x))
        _y = ffi.cast("float *", ffi.from_buffer(y))
        NervanaObject.be.mathlib.cmath_axpby(x.size, a, _x, b, _y)
    else:
        y = a*x + b*y


def change_data_store_order(a, a_row, a_col, a_len, b, axis=1):
    if iscontiguous(a._tensor):
        _a = ffi.cast("float *", ffi.from_buffer(a._tensor))
        _b = ffi.cast("float *", ffi.from_buffer(b._tensor))
        NervanaObject.be.mathlib.cmath_change_data_store_order(_a, axis, a_row, a_col, a_len, _b)

        return b
    else:
        raise AttributeError('a should be contiguous')


def add_and_act(h, h_ff, bias, activation):
    if iscontiguous(h) and iscontiguous(h_ff) and iscontiguous(bias):
        if isinstance(activation, Rectlinclip) and (activation.slope == 0):
            _h = ffi.cast("float *", ffi.from_buffer(h))
            _h_ff = ffi.cast("float *", ffi.from_buffer(h_ff))
            _bias = ffi.cast("float *", ffi.from_buffer(bias))
            NervanaObject.be.mathlib.cmath_add_and_act(_h, _h_ff, _bias, h.shape[0], h.shape[1],
                                                       activation.xcut)
            return True
        else:
            return False
    else:
        return False


def trans2d(W_recur_f, W_recur_b, W_recur_f_contiguous, W_recur_b_contiguous):
    if iscontiguous(W_recur_f._tensor) and iscontiguous(W_recur_b._tensor) \
       and W_recur_f.shape == W_recur_b.shape:
        _in_0 = ffi.cast("float *", ffi.from_buffer(W_recur_f._tensor))
        _in_1 = ffi.cast("float *", ffi.from_buffer(W_recur_b._tensor))
        _out_0 = ffi.cast("float *", ffi.from_buffer(W_recur_f_contiguous._tensor))
        _out_1 = ffi.cast("float *", ffi.from_buffer(W_recur_b_contiguous._tensor))
        NervanaObject.be.mathlib.cmath_trans2d(_in_0, _in_1, W_recur_f.shape[0],
                                               W_recur_f.shape[1], _out_0, _out_1)


def act_and_mul(in_deltas, hs, activation):
    if iscontiguous(in_deltas._tensor) and iscontiguous(hs._tensor):
        if isinstance(activation, Rectlinclip) and (activation.slope == 0):
            _in_deltas = ffi.cast("float *", ffi.from_buffer(in_deltas._tensor))
            _hs = ffi.cast("float *", ffi.from_buffer(hs._tensor))
            NervanaObject.be.mathlib.cmath_act_and_mul(_in_deltas, _hs, in_deltas.size,
                                                       activation.xcut)
            return True
        else:
            return False
    else:
        return False



# ----------------------------------------------------------------------------
# Copyright 2015-2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from functools import wraps

from neon import NervanaObject
from collections import OrderedDict
from neon import __version__ as __neon_version__
from neon import logger as neon_logger
from neon.layers.container import LayerContainer
import numpy as np


@property
def flatten_layers(self):
    lf = []
    for l in self.layers:
        if isinstance(l, LayerContainer):
            lf += l.flatten_layers
        else:
            lf.append(l)
    return lf


class Benchmark(object):
    def __init__(self, model, time_layers=True):
        self.model = model
        self.layer_times = OrderedDict()
        if not time_layers:
            return

        self.time_layers = time_layers
        self.start = self.model.be.init_mark()
        self.end = self.model.be.init_mark()
        LayerContainer.flatten_layers = flatten_layers
        for idx, layer in enumerate(self.model.layers.flatten_layers):
            l_name = layer.name

            f_l_name = 'fprop ' + l_name
            b_l_name = 'bprop ' + l_name
            self.layer_times[f_l_name] = []
            self.layer_times[b_l_name] = []
            layer.fprop = Benchmark.benchmarked(layer.fprop, self.start,
                                                self.end, self.layer_times[f_l_name])
            layer.bprop = Benchmark.benchmarked(layer.bprop, self.start,
                                                self.end, self.layer_times[b_l_name])

    @staticmethod
    def benchmarked(func, start, end, output):
        '''
        :param func: Function that is timed
        :param start: Start marker
        :param end: End marker
        :param output: Array to which func execution time will be appended
        :return: Wrapped function
        '''
        nerv_be = NervanaObject.be

        @wraps(func)
        def wrapper(*args, **kwargs):
            nerv_be.record_mark(start)
            res = func(*args, **kwargs)
            nerv_be.record_mark(end)
            nerv_be.synchronize_mark(end)
            output.append(nerv_be.get_time(start, end))
            return res

        return wrapper

    def time(self, dataset, inference=False, niterations=20):
        """
        Measure runtime for computing fprop and bprop separately, as well as
        full minibatch run times. For inference case, only the fprop is measured.

        Arguments:
             dataset (NervanaDataIterator) Dataset iterator to perform fit on

             inference (bool, optional): Is inference use case

             niterations (optional, int): Number of minibatches to average over
        Returns:
            dictionary with fprop, bprop run times
        """
        if not dataset:
            raise ValueError("Provide correct dataset for timing")

        if niterations < 1:
            raise ValueError("Invalid iterations number")

        if inference is False and (self.model.cost is None or self.model.optimizer is None):
            raise RuntimeError("Need cost or optimizer to benchmark bprop")

        # iterate through minibatches of the dataset
        times = OrderedDict()
        time_keys = ['fprop'] if inference else ['fprop', 'bprop', 'iteration']
        if self.time_layers:
            time_keys.append('data_loading')
        for ky in time_keys:
            times[ky] = np.full(niterations, -1.0)
        count = 0

        data_loading_mark = self.model.be.init_mark()
        fprop_start = self.model.be.init_mark()
        fprop_end = self.model.be.init_mark()
        bprop_end = self.model.be.init_mark()

        while count < niterations:
            dataset.reset()
            self.model.be.record_mark(data_loading_mark)
            for _, (x, t) in enumerate(dataset):

                self.model.be.record_mark(fprop_start)  # mark start of fprop
                x = self.model.fprop(x, inference)
                self.model.be.record_mark(fprop_end)  # mark end of fprop and start of bprop

                if inference is False:
                    self.model.cost.get_cost(x, t)
                    delta = self.model.cost.get_errors(x, t)
                    self.model.bprop(delta)
                    self.model.optimizer.optimize(self.model.layers_to_optimize, epoch=0)

                    self.model.be.record_mark(bprop_end)  # mark end of bprop
                    self.model.be.synchronize_mark(bprop_end)
                else:
                    self.model.be.synchronize_mark(fprop_end)

                data_loading_time = self.model.be.get_time(data_loading_mark, fprop_start)
                if self.time_layers:
                    times['data_loading'][count] = data_loading_time
                times['fprop'][count] = self.model.be.get_time(fprop_start,
                                                               fprop_end) + data_loading_time
                if inference is False:
                    times['bprop'][count] = self.model.be.get_time(fprop_end, bprop_end)
                    times['iteration'][count] = times['fprop'][
                        count] + times['bprop'][count]

                self.model.be.record_mark(data_loading_mark)

                count += 1
                if count >= niterations:
                    break

        times.update(self.layer_times)
        return times

    def time_gan(self, dataset, niterations=101):
        """
        Measure runtime by computing fprop and bprop times on GAN networks

        Arguments:
             dataset (NervanaDataIterator) Dataset iterator to perform fit on

             niterations (optional, int): Number of minibatches to average over
        Returns:
            dictionary with fprop, bprop run times
        """

        if not dataset:
            raise ValueError("Provide correct dataset for timing")

        min_iterations = self.model.get_k(self.model.gen_iter) + 1
        if niterations < min_iterations:
            raise ValueError("Invalid iterations number. Run at least " +
                             str(min_iterations) + " iterations.")

        # iterate through minibatches of the dataset
        times = OrderedDict()
        time_keys = ['fprop', 'bprop', 'iteration']
        if self.time_layers:
            time_keys.append('data_loading')
        for ky in time_keys:
            times[ky] = np.full(niterations, -1.0)
        count = 0

        epoch = self.model.epoch_index
        z, y_temp = self.model.zbuf, self.model.ybuf

        data_loading_mark = self.model.be.init_mark()
        fprop_start = self.model.be.init_mark()
        fprop_end = self.model.be.init_mark()
        bprop_start = self.model.be.init_mark()
        bprop_end = self.model.be.init_mark()

        while count < niterations:
            dataset.reset()
            self.model.be.record_mark(data_loading_mark)
            for mb_idx, (x, t) in enumerate(dataset):
                # clip all discriminator parameters to a cube in case of WGAN
                if self.model.wgan_param_clamp:
                    self.model.clip_param_in_layers(
                        self.model.layers.discriminator.layers_to_optimize,
                        self.model.wgan_param_clamp)
                # benchmark discriminator on noise
                self.model.fill_noise(z, normal=(self.model.noise_type == 'normal'))
                self.model.be.record_mark(fprop_start)  # mark start of fprop
                Gz = self.model.fprop_gen(z)
                y_noise = self.model.fprop_dis(Gz)
                self.model.be.record_mark(fprop_end)  # mark end of fprop
                data_loading_time = self.model.be.get_time(data_loading_mark, fprop_start)
                if self.time_layers:
                    times['data_loading'][count] = data_loading_time
                times['fprop'][count] += self.model.be.get_time(fprop_start,
                                                                fprop_end) + data_loading_time
                y_temp[:] = y_noise
                self.model.be.record_mark(bprop_start)  # mark start of bprop
                delta_noise = self.model.cost.costfunc.bprop_noise(y_noise)
                self.model.bprop_dis(delta_noise)
                self.model.be.record_mark(bprop_end)  # mark end of bprop
                times['bprop'][count] += self.model.be.get_time(bprop_start, bprop_end)
                self.model.layers.discriminator.set_acc_on(True)

                # benchmark discriminator on data
                self.model.be.record_mark(fprop_start)
                y_data = self.model.fprop_dis(x)
                self.model.be.record_mark(fprop_end)
                times['fprop'][count] += self.model.be.get_time(fprop_start, fprop_end)
                self.model.be.record_mark(bprop_start)
                delta_data = self.model.cost.costfunc.bprop_data(y_data)
                self.model.bprop_dis(delta_data)
                self.model.optimizer.optimize(self.model.layers.discriminator.layers_to_optimize,
                                              epoch=epoch)
                self.model.be.record_mark(bprop_end)
                times['bprop'][count] += self.model.be.get_time(bprop_start, bprop_end)
                self.model.layers.discriminator.set_acc_on(False)

                # benchmark generator
                if self.model.current_batch == self.model.last_gen_batch + \
                        self.model.get_k(self.model.gen_iter):
                    self.model.fill_noise(z, normal=(self.model.noise_type == 'normal'))
                    self.model.be.record_mark(fprop_start)
                    Gz = self.model.fprop_gen(z)
                    y_temp[:] = y_data
                    y_noise = self.model.fprop_dis(Gz)
                    self.model.be.record_mark(fprop_end)
                    times['fprop'][count] += self.model.be.get_time(fprop_start, fprop_end)
                    self.model.be.record_mark(bprop_start)
                    delta_noise = self.model.cost.costfunc.bprop_generator(y_noise)
                    delta_dis = self.model.bprop_dis(delta_noise)
                    self.model.bprop_gen(delta_dis)
                    self.model.optimizer.optimize(self.model.layers.generator.layers_to_optimize,
                                                  epoch=epoch)
                    self.model.be.record_mark(bprop_end)
                    times['bprop'][count] += self.model.be.get_time(bprop_start, bprop_end)
                    self.model.last_gen_batch = self.model.current_batch
                    self.model.gen_iter += 1

                self.model.current_batch += 1
                times['iteration'][count] = times['data_loading'][count] + times['fprop'][
                     count] + times['bprop'][count]
                self.model.be.record_mark(data_loading_mark)
                count += 1
                if count >= niterations:
                    break

        return times

    @staticmethod
    def print_stats(stats, functions=None, output=None, nskip=0):
        '''
        Function prints provided statistics
        :param stats: Dictionary with statistics
        :param functions: Functions to be executed on statistics
        :param output: Object which provides display method for printing
        :param nskip: Number of measurements to be skipped in functions
        '''
        if not stats:
            raise ValueError('Provide stats')

        if not output:
            output = neon_logger
        elif not hasattr(output, 'display'):
            raise TypeError('Argument output must implement display method')

        if not functions:
            functions = (np.mean, np.median, np.min, np.max)

        function_names = tuple(function.__name__.title() for function in functions)

        fmt_titles = '| {name:^33} '.format(name='Func') + '| {:^11} ' * len(function_names) \
                     + '|    Units    |'
        fmt_nums = '| {func:<33} ' + '|  {%s:<10.5g} ' * len(function_names) % function_names \
                   + '| {units:^11} |'

        head_str = fmt_titles.format(*function_names)
        sep = '+' + '-' * (len(head_str) - 2) + '+'

        msg = '\n{sep:}\n|{prefix:-^105}|\n' \
            .format(sep=sep, prefix='  Intel Neon Benchmark: 0.0.2 Neon Version: %s  '
                                    % __neon_version__[:5])
        msg += '%s\n%s\n%s\n' % (sep, head_str, sep)
        out_stats = {}
        for step in stats:
            if not any(stats[step]):
                continue
            timesu = np.array(stats[step][nskip:])  # in ms
            out_stats[step] = {}
            for function in functions:
                out_stats[step][function.__name__.title()] = function(timesu)
            msg += '%s\n' % fmt_nums.format(units='msec', func=step, **out_stats[step])
        msg += sep

        output.display(msg)



from pycuda.tools import context_dependent_memoize

# from neon.backends.cuda_templates import (_ew_template,
#                                           _stage_template,
#                                           _fin_template,
#                                           _init_rand_func,
#                                           _init_rand_round_func,
#                                           _finish_rand_func,
#                                           _common_urand_gen,
#                                           _common_frand,
#                                           _common_round,
#                                           _common_fp16_to_fp32,
#                                           _ew_types,
#                                           _ew_strings,
#                                           _is_finite,
#                                           _float_ops,
#                                           _reduction_ops)
from neon.backends.cuda_templates import (_common_round,
                                          _common_kepler,
                                          _ew_types,
                                          _common_fp16_to_fp32,
                                          _ew_strings)

from neon.backends.kernels.cuda.binary import shift_element
from neon.backends.util.source_module import SourceModule

@context_dependent_memoize
def _get_bn_fprop_kernel(dtype, threads, compute_capability):

    if threads > 32:
        shr_code = "__shared__ float sPartials[THREADS];"
        red_code = r"""
    sPartials[tid] = xvar;
    __syncthreads();

    #pragma unroll
    for (int a = THREADS >> 1; a > 32; a >>= 1)
    {
        if ( tid < a )
            sPartials[tid] += sPartials[tid + a];
        __syncthreads();
    }
    if ( tid < 32 )
    {
        xvar = sPartials[tid] + sPartials[tid + 32];
        #pragma unroll
        for (int i = 16; i > 0; i >>= 1)
            xvar += __shfl_xor(xvar, i);

        sPartials[tid] = xvar * rcpN;
    }
    __syncthreads();
    xvar = sPartials[0];
"""
    else:
        shr_code = ""
        red_code = r"""
    #pragma unroll
    for (int i = 16; i > 0; i >>= 1)
        xvar += __shfl_xor(xvar, i);
    xvar *= rcpN;
"""

    code = r"""
#define THREADS %(threads)s

%(common)s
%(binary)s

__global__ void batchnorm_fprop (
    %(type)s* y_out, float* xvar_out, float* gmean_out, float* gvar_out,
    const %(type)s* x_in, const float* xsum_in, const float* gmean_in,
    const float* gvar_in, const float* gamma_in, const float* beta_in,
    const float eps, const float rho, const float accumbeta, const int N,
    const int relu, bool binary)
{
    %(share)s

    const int tid  = threadIdx.x;
    const int bid  = blockIdx.x;
    int offset = bid * N;

    const %(type)s* x_in0 = x_in + offset + tid;

    const float rcpN = 1.0f/(float)N;

    float xmean = __ldg(xsum_in + bid) * rcpN;

    float xvar = 0.0f;
    for (int i = tid; i < N; i += THREADS)
    {
        float x = %(cvt)s(__ldg(x_in0));
        x_in0 += THREADS;

        x -= xmean;
        if (binary) {
            xvar += shift_element(x, x, true);
        } else {
            xvar += x * x;
        }
    }
    %(red)s

    float gamma = __ldg(gamma_in + bid);
    float beta  = __ldg(beta_in  + bid);

    if ( tid == 0 )
    {
        float gmean = __ldg(gmean_in + bid);
        float gvar  = __ldg(gvar_in  + bid);

        *(xvar_out  + bid) = xvar;
        *(gmean_out + bid) = gmean * rho + (1.0f - rho) * xmean;
        *(gvar_out  + bid) = gvar  * rho + (1.0f - rho) * xvar;
    }

    float xvar_rcp_sqrt = 1.0f / sqrtf(xvar + eps);

    int start = N - (THREADS*4 - tid);
    offset += start;
    x_in   += offset;
    y_out  += offset;

    for (int i = start; i >= -THREADS*3; i -= THREADS*4)
    {
        float x0 = i >= -THREADS*0 ? %(cvt)s(__ldg(x_in + THREADS*0)) : 0.0f;
        float x1 = i >= -THREADS*1 ? %(cvt)s(__ldg(x_in + THREADS*1)) : 0.0f;
        float x2 = i >= -THREADS*2 ? %(cvt)s(__ldg(x_in + THREADS*2)) : 0.0f;
        float x3 =                   %(cvt)s(__ldg(x_in + THREADS*3));

        x_in -= THREADS*4;

        float xhat0 = 0.0f;
        float xhat1 = 0.0f;
        float xhat2 = 0.0f;
        float xhat3 = 0.0f;

        float y0 = 0.0f;
        float y1 = 0.0f;
        float y2 = 0.0f;
        float y3 = 0.0f;
        if (binary) {
            xhat0 = shift_element(x0 - xmean, xvar_rcp_sqrt, true);
            xhat1 = shift_element(x1 - xmean, xvar_rcp_sqrt, true);
            xhat2 = shift_element(x2 - xmean, xvar_rcp_sqrt, true);
            xhat3 = shift_element(x3 - xmean, xvar_rcp_sqrt, true);

            y0 = shift_element(xhat0, gamma, true) + beta;
            y1 = shift_element(xhat1, gamma, true) + beta;
            y2 = shift_element(xhat2, gamma, true) + beta;
            y3 = shift_element(xhat3, gamma, true) + beta;
        } else {
            xhat0 = (x0 - xmean) * xvar_rcp_sqrt;
            xhat1 = (x1 - xmean) * xvar_rcp_sqrt;
            xhat2 = (x2 - xmean) * xvar_rcp_sqrt;
            xhat3 = (x3 - xmean) * xvar_rcp_sqrt;

            y0 = xhat0 * gamma + beta;
            y1 = xhat1 * gamma + beta;
            y2 = xhat2 * gamma + beta;
            y3 = xhat3 * gamma + beta;
        }

        if (relu)
        {
            y0 = fmaxf(y0, 0.0f);
            y1 = fmaxf(y1, 0.0f);
            y2 = fmaxf(y2, 0.0f);
            y3 = fmaxf(y3, 0.0f);
        }

        %(y0_out)s
        %(y1_out)s
        %(y2_out)s
        %(y3_out)s
        if (accumbeta == 0.0)
        {
            if (i >= -THREADS*0) *(y_out + THREADS*0) = y0_val;
            if (i >= -THREADS*1) *(y_out + THREADS*1) = y1_val;
            if (i >= -THREADS*2) *(y_out + THREADS*2) = y2_val;
                                 *(y_out + THREADS*3) = y3_val;
        }
        else
        {
            if (i >= -THREADS*0) *(y_out + THREADS*0) = y_out[THREADS*0] * accumbeta + y0_val;
            if (i >= -THREADS*1) *(y_out + THREADS*1) = y_out[THREADS*1] * accumbeta + y1_val;
            if (i >= -THREADS*2) *(y_out + THREADS*2) = y_out[THREADS*2] * accumbeta + y2_val;
                                 *(y_out + THREADS*3) = y_out[THREADS*3] * accumbeta + y3_val;
        }
        y_out -= THREADS*4;
    }
}
"""
    out_code = _ew_strings["round"]["nearest"].get(dtype, "float {0} = {1};")
    common_code  = _common_round["nearest"].get(dtype, "")
    if dtype == "f2":
        common_code += _common_fp16_to_fp32

    if (compute_capability[0] == 3 and compute_capability[1] < 5) or compute_capability[0] < 3:
        common_code += _common_kepler

    code = code % {
        "common"    : common_code,
        "binary"    : shift_element(),
        "share"     : shr_code,
        "red"       : red_code,
        "threads"   : threads,
        "type"      : _ew_types[dtype]["type"],
        "cvt"       : _ew_types[dtype]["cvt"],
        "y0_out"    : out_code.format("y0_val",     "y0"),
        "y1_out"    : out_code.format("y1_val",     "y1"),
        "y2_out"    : out_code.format("y2_val",     "y2"),
        "y3_out"    : out_code.format("y3_val",     "y3"),
    }
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("batchnorm_fprop")
    kernel.prepare("PPPPPPPPPPfffIII")
    kernel.name = "batchnorm_fprop"
    return kernel


@context_dependent_memoize
def _get_bn_bprop_kernel(dtype, threads, compute_capability):

    if threads > 32:
        shr_code = "__shared__ float sPartials[THREADS * 2];"
        red_code = r"""
    sPartials[tid + THREADS*0] = grad_gamma;
    sPartials[tid + THREADS*1] = grad_beta;
    __syncthreads();

    #pragma unroll
    for (int a = THREADS >> 1; a > 32; a >>= 1)
    {
        if ( tid < a )
        {
            sPartials[tid + THREADS*0] += sPartials[tid + a + THREADS*0];
            sPartials[tid + THREADS*1] += sPartials[tid + a + THREADS*1];
        }
        __syncthreads();
    }
    if ( tid < 32 )
    {
        grad_gamma = sPartials[tid + THREADS*0] + sPartials[tid + 32 + THREADS*0];
        grad_beta  = sPartials[tid + THREADS*1] + sPartials[tid + 32 + THREADS*1];

        #pragma unroll
        for (int i = 16; i > 0; i >>= 1)
        {
            grad_gamma += __shfl_xor(grad_gamma, i);
            grad_beta  += __shfl_xor(grad_beta,  i);
        }
        sPartials[tid + THREADS*0] = grad_gamma;
        sPartials[tid + THREADS*1] = grad_beta;
    }
    __syncthreads();
    grad_gamma = sPartials[THREADS*0];
    grad_beta  = sPartials[THREADS*1];
"""
    else:
        shr_code = ""
        red_code = r"""
    #pragma unroll
    for (int i = 16; i > 0; i >>= 1)
    {
        grad_gamma += __shfl_xor(grad_gamma, i);
        grad_beta  += __shfl_xor(grad_beta,  i);
    }
"""

    code = r"""
#define THREADS %(threads)s

%(common)s
%(binary)s

__global__ void batchnorm_bprop (
    %(type)s* delta_out, float* grad_gamma_out, float* grad_beta_out,
    const %(type)s* delta_in, const %(type)s* x_in, const float* xsum_in,
    const float* xvar_in, const float* gamma_in,
    const float eps, const int N, bool binary)
{
    %(share)s

    const int tid  = threadIdx.x;
    const int bid  = blockIdx.x;
    const float rcpN = 1.0f/(float)N;
    int offset = bid * N;

    const %(type)s* x_in0 = x_in     + offset + tid;
    const %(type)s* d_in0 = delta_in + offset + tid;

    float xmean = __ldg(xsum_in  + bid) * rcpN;
    float xvar  = __ldg(xvar_in  + bid);
    float gamma = __ldg(gamma_in + bid);

    float xvar_rcp_sqrt = 1.0f / sqrtf(xvar + eps);
    float grad_gamma    = 0.0f;
    float grad_beta     = 0.0f;

    for (int i = tid; i < N; i += THREADS)
    {
        float x = %(cvt)s(__ldg(x_in0));
        x_in0 += THREADS;
        float d = %(cvt)s(__ldg(d_in0));
        d_in0 += THREADS;

        float xhat = 0.0f;
        if (binary) {
            xhat = shift_element(x - xmean, xvar_rcp_sqrt, true);
        } else {
            xhat = (x - xmean) * xvar_rcp_sqrt;
        }

        grad_gamma += xhat * d;
        grad_beta  += d;
    }
    %(red)s

    if ( tid == 0 )
    {
        *(grad_gamma_out + bid) = grad_gamma;
        *(grad_beta_out  + bid) = grad_beta;
    }

    int start = N - (THREADS*4 - tid);
    offset += start;
    const %(type)s* x_in1 = x_in     + offset;
    const %(type)s* d_in1 = delta_in + offset;
    delta_out += offset;

    for (int i = start; i >= -THREADS*3; i -= THREADS*4)
    {
        float x0 = i >= -THREADS*0 ? %(cvt)s(__ldg(x_in1 + THREADS*0)) : 0.0f;
        float x1 = i >= -THREADS*1 ? %(cvt)s(__ldg(x_in1 + THREADS*1)) : 0.0f;
        float x2 = i >= -THREADS*2 ? %(cvt)s(__ldg(x_in1 + THREADS*2)) : 0.0f;
        float x3 =                   %(cvt)s(__ldg(x_in1 + THREADS*3));

        float d0 = i >= -THREADS*0 ? %(cvt)s(__ldg(d_in1 + THREADS*0)) : 0.0f;
        float d1 = i >= -THREADS*1 ? %(cvt)s(__ldg(d_in1 + THREADS*1)) : 0.0f;
        float d2 = i >= -THREADS*2 ? %(cvt)s(__ldg(d_in1 + THREADS*2)) : 0.0f;
        float d3 =                   %(cvt)s(__ldg(d_in1 + THREADS*3));

        x_in1 -= THREADS*4;
        d_in1 -= THREADS*4;

        float xhat0 = 0.0f;
        float xhat1 = 0.0f;
        float xhat2 = 0.0f;
        float xhat3 = 0.0f;

        float xtmp0 = 0.0f;
        float xtmp1 = 0.0f;
        float xtmp2 = 0.0f;
        float xtmp3 = 0.0f;

        float delta0 = 0.0f;
        float delta1 = 0.0f;
        float delta2 = 0.0f;
        float delta3 = 0.0f;

        if (binary) {
            xhat0 = shift_element(x0 - xmean, xvar_rcp_sqrt, true);
            xhat1 = shift_element(x1 - xmean, xvar_rcp_sqrt, true);
            xhat2 = shift_element(x2 - xmean, xvar_rcp_sqrt, true);
            xhat3 = shift_element(x3 - xmean, xvar_rcp_sqrt, true);

            xtmp0 = (shift_element(xhat0, grad_gamma, true) + grad_beta) * rcpN;
            xtmp1 = (shift_element(xhat1, grad_gamma, true) + grad_beta) * rcpN;
            xtmp2 = (shift_element(xhat2, grad_gamma, true) + grad_beta) * rcpN;
            xtmp3 = (shift_element(xhat3, grad_gamma, true) + grad_beta) * rcpN;

            delta0 = shift_element(shift_element(d0 - xtmp0, gamma, true), xvar_rcp_sqrt, true);
            delta1 = shift_element(shift_element(d1 - xtmp1, gamma, true), xvar_rcp_sqrt, true);
            delta2 = shift_element(shift_element(d2 - xtmp2, gamma, true), xvar_rcp_sqrt, true);
            delta3 = shift_element(shift_element(d3 - xtmp3, gamma, true), xvar_rcp_sqrt, true);
        } else {
            xhat0 = (x0 - xmean) * xvar_rcp_sqrt;
            xhat1 = (x1 - xmean) * xvar_rcp_sqrt;
            xhat2 = (x2 - xmean) * xvar_rcp_sqrt;
            xhat3 = (x3 - xmean) * xvar_rcp_sqrt;

            xtmp0 = (xhat0 * grad_gamma + grad_beta) * rcpN;
            xtmp1 = (xhat1 * grad_gamma + grad_beta) * rcpN;
            xtmp2 = (xhat2 * grad_gamma + grad_beta) * rcpN;
            xtmp3 = (xhat3 * grad_gamma + grad_beta) * rcpN;

            delta0 = gamma * (d0 - xtmp0) * xvar_rcp_sqrt;
            delta1 = gamma * (d1 - xtmp1) * xvar_rcp_sqrt;
            delta2 = gamma * (d2 - xtmp2) * xvar_rcp_sqrt;
            delta3 = gamma * (d3 - xtmp3) * xvar_rcp_sqrt;
        }

        %(delta0_out)s
        %(delta1_out)s
        %(delta2_out)s
        %(delta3_out)s
        if (i >= -THREADS*0) *(delta_out + THREADS*0) = delta0_val;
        if (i >= -THREADS*1) *(delta_out + THREADS*1) = delta1_val;
        if (i >= -THREADS*2) *(delta_out + THREADS*2) = delta2_val;
                             *(delta_out + THREADS*3) = delta3_val;
        delta_out -= THREADS*4;
    }
}
"""
    out_code = _ew_strings["round"]["nearest"].get(dtype, "float {0} = {1};")
    common_code = _common_round["nearest"].get(dtype, "")
    if dtype == "f2":
        common_code += _common_fp16_to_fp32

    if (compute_capability[0] == 3 and compute_capability[1] < 5) or compute_capability[0] < 3:
        common_code += _common_kepler

    code = code % {
        "common"         : common_code,
        "binary"         : shift_element(),
        "share"          : shr_code,
        "red"            : red_code,
        "threads"        : threads,
        "type"           : _ew_types[dtype]["type"],
        "cvt"            : _ew_types[dtype]["cvt"],
        "delta0_out"     : out_code.format("delta0_val",     "delta0"),
        "delta1_out"     : out_code.format("delta1_val",     "delta1"),
        "delta2_out"     : out_code.format("delta2_val",     "delta2"),
        "delta3_out"     : out_code.format("delta3_val",     "delta3"),
    }
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("batchnorm_bprop")
    kernel.prepare("PPPPPPPPfII")
    kernel.name = "batchnorm_bprop"
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

"""
Automatic differentiation of optrees
Supports:
    - elementwise operations
      - unary
      - binary
    - dot
    - reductions (<2d, same as nervanagpu)
To support:
    - batched_dot
    - zero-operand operations
    - slicing (need to modify tensor view)
TODO:
    - make use of empty_like
    - intrinsic key for caching
"""
from __future__ import division
from builtins import object, zip
from neon.backends.backend import OpTreeNode, Tensor
import numpy as np
from functools import wraps

_scalar_types = {int, float, np.float16, np.float32, np.uint8, np.int8,
                 np.uint16, np.int16, np.uint32, np.int32}


class GradUtil(object):

    """
    Utility class for calculating gradients.
    """

    @staticmethod
    def get_grad_back(grad_node):
        """
        Get left and right gradient increments from back-propagation.

        Arguments:
            grad_node (GradNode): The GradNode to perform gradient
                                  back-propagation on.
        """
        if not grad_node:
            return None
        # dissemble node
        x = grad_node.left.op_tree if grad_node.left else None
        y = grad_node.right.op_tree if grad_node.right else None
        z = grad_node.op_tree
        dz = grad_node.grad_op_tree
        op_dict = z[0]
        be = grad_node.ad.be
        # get element-wise gradient increments
        op = grad_node.op_tree[0]['op']
        grad_increments = grad_map[op](x, y, z, dz, op_dict, be)
        # unbroacat to input dimensions
        left_increment = GradUtil._unbroadcast(grad_increments[0], x, be)
        right_increment = GradUtil._unbroadcast(grad_increments[1], y, be)

        return (left_increment, right_increment)

    @staticmethod
    def _unbroadcast(grad_op_tree, x, be):
        """
        Reverse broadcast from shape(grad_op_tree) to shape(x)

        Arguments:
            grad_op_tree (OpTreeNode or Tensor): The OpTreeNode to broadcast.
            x (OpTreeNode or Tensor): Provides the dimension to be broadcasted to.
            be: (Backend): The backend to be used.

        Returns:
            OpTreeNode or Tensor: The broadcasted result.
        """

        if (not grad_op_tree) or (not x):
            return grad_op_tree
        if type(x) in _scalar_types:
            return 0.
        in_shape = x.shape
        out_shape = grad_op_tree.shape

        if in_shape == out_shape:
            return grad_op_tree
        elif len(in_shape) == 2 and len(out_shape) == 2:
            # broadcasts
            if in_shape == (1, 1):
                # [1 * 1] -> [m * n]
                return be.sum(grad_op_tree)
            elif in_shape[0] == out_shape[0] and in_shape[1] == 1:
                # [m * 1] -> [m * n]
                return be.sum(grad_op_tree, axis=1)
            elif in_shape[0] == 1 and in_shape[1] == out_shape[1]:
                # [1 * n] -> [m * n]
                return be.sum(grad_op_tree, axis=0)
            # reductions
            elif ((out_shape[0] == in_shape[0] and out_shape[1] == 1) or
                  (out_shape[0] == 1 and out_shape[1] == in_shape[1])):
                # TODO cleaner way to broadcast
                return 0 * x + grad_op_tree
            else:
                return NotImplemented
        else:
            return NotImplemented

    @staticmethod
    def is_invalid(grad_op_tree, be):
        """
        Test if the result of grad_op_tree contains Nan, inf, -inf, or
        abnormally large or small numbers. Only for debug purpose.

        Arguments:
            grad_op_tree (OpTreeNode or Tensor): The tensor or op-tree to test.
            be (Backend): The backend to be used.

        Returns:
            bool: Whether the result contains Nan, inf, -inf, or abnormally
                  large or small numbers
        """
        grad_op_tree_val = be.empty(grad_op_tree.shape)
        grad_op_tree_val[:] = grad_op_tree
        grad_op_tree_val_np = grad_op_tree_val.get().reshape(-1,)
        for val in grad_op_tree_val_np:
            if not (-50000. < val < 50000.):
                return True
        else:
            return False

    """
    (Applies to the following grad functions)
    Return gradients for these operations.

    Arguments:
        x (Tensor, int, float, OpTreeNode): Left operand.
        y (Tensor, int, float, OpTreeNode): Right operand.
        z (Tensor, int, float, OpTreeNode): `z = x op y`
        dz (Tensor, int, float, OpTreeNode): Gradient w.r.t.`z`
        op_dict (dict): Dictionary specifying the operation.
        be (Backend): The backend of the tensors.

    Returns:
        Tuple: (left_increment, right_increment)
    """

    # derivatives
    @staticmethod
    def _zero_grad_unary(x, y, z, dz, op_dict, be):
        return (dz * 0., None)

    @staticmethod
    def _zero_grad_binary(x, y, z, dz, op_dict, be):
        return (dz * 0., dz * 0.)

    @staticmethod
    def _add_grad(x, y, z, dz, op_dict, be):
        return (dz, dz)

    @staticmethod
    def _mul_grad(x, y, z, dz, op_dict, be):
        return (dz * y, dz * x)

    @staticmethod
    def _sub_grad(x, y, z, dz, op_dict, be):
        return (dz, -dz)

    @staticmethod
    def _neg_grad(x, y, z, dz, op_dict, be):
        return (-dz, None)

    @staticmethod
    def _pow_grad(x, y, z, dz, op_dict, be):
        return (dz * y * x ** (y - 1.), dz * z * be.log(x))

    @staticmethod
    def _div_grad(x, y, z, dz, op_dict, be):
        return (dz / y, -dz * x / be.square(y))

    @staticmethod
    def _dot_grad(x, y, z, dz, op_dict, be):
        return (be.dot(dz, y.T), be.dot(x.T, dz))

    @staticmethod
    def _abs_grad(x, y, z, dz, op_dict, be):
        return (dz * be.sgn(x), None)

    @staticmethod
    def _sqrt_grad(x, y, z, dz, op_dict, be):
        return (dz * 0.5 / z, None)

    @staticmethod
    def _sqr_grad(x, y, z, dz, op_dict, be):
        return (dz * 2. * x, None)

    @staticmethod
    def _exp_grad(x, y, z, dz, op_dict, be):
        return (dz * z, None)

    @staticmethod
    def _exp2_grad(x, y, z, dz, op_dict, be):
        return (dz * z * be.log(2.), None)

    @staticmethod
    def _log_grad(x, y, z, dz, op_dict, be):
        return (dz / x, None)

    @staticmethod
    def _log2_grad(x, y, z, dz, op_dict, be):
        return (dz / x / be.log(2.), None)

    @staticmethod
    def _sig_grad(x, y, z, dz, op_dict, be):
        return (dz * z * (1. - z), None)

    @staticmethod
    def _sig2_grad(x, y, z, dz, op_dict, be):
        return (dz * z * (1. - z) * be.log(2.), None)

    @staticmethod
    def _tanh_grad(x, y, z, dz, op_dict, be):
        return (dz * (1. - be.square(z)), None)

    @staticmethod
    def _tanh2_grad(x, y, z, dz, op_dict, be):
        return (dz * (1. - be.square(z)) * be.log(2.), None)

    @staticmethod
    def _max_grad(x, y, z, dz, op_dict, be):
        return (dz * (x == z), None)

    @staticmethod
    def _min_grad(x, y, z, dz, op_dict, be):
        return (dz * (x == z), None)

    @staticmethod
    def _maximum_grad(x, y, z, dz, op_dict, be):
        return (dz * be.greater_equal(x, y), dz * be.greater_equal(y, x))

    @staticmethod
    def _minimum_grad(x, y, z, dz, op_dict, be):
        return (dz * be.less_equal(x, y), dz * be.less_equal(y, x))

    @staticmethod
    def _sum_grad(x, y, z, dz, op_dict, be):
        assert('axis' in op_dict and (op_dict['axis'] in (0, 1)))
        return (dz, None)  # will be unbroadcasted

    @staticmethod
    def _transpose_grad(x, y, z, dz, op_dict, be):
        return (dz.T, None)


grad_map = {
    # zero gradients
    'eq': GradUtil._zero_grad_binary,
    'lt': GradUtil._zero_grad_binary,
    'le': GradUtil._zero_grad_binary,
    'gt': GradUtil._zero_grad_binary,
    'ge': GradUtil._zero_grad_binary,
    'sgn': GradUtil._zero_grad_unary,
    'finite': GradUtil._zero_grad_unary,
    'argmax': GradUtil._zero_grad_unary,
    'argmin': GradUtil._zero_grad_unary,
    # binary operations
    'add': GradUtil._add_grad,
    'mul': GradUtil._mul_grad,
    'sub': GradUtil._sub_grad,
    'pow': GradUtil._pow_grad,
    'div': GradUtil._div_grad,
    'dot': GradUtil._dot_grad,
    # unary operations
    'neg': GradUtil._neg_grad,
    'abs': GradUtil._abs_grad,
    'sqrt': GradUtil._sqrt_grad,
    'sqr': GradUtil._sqr_grad,
    'exp': GradUtil._exp_grad,
    'exp2': GradUtil._exp2_grad,
    'log': GradUtil._log_grad,
    'log2': GradUtil._log2_grad,
    'sig': GradUtil._sig_grad,
    'sig2': GradUtil._sig2_grad,
    'tanh': GradUtil._tanh_grad,
    'tanh2': GradUtil._tanh2_grad,
    'max': GradUtil._max_grad,
    'min': GradUtil._min_grad,
    'maximum': GradUtil._maximum_grad,
    'minimum': GradUtil._minimum_grad,
    # reduction operations
    'sum': GradUtil._sum_grad,
    # transpose
    'transpose': GradUtil._transpose_grad
}


def memoize_autodiff(func):
    """
    Memoize to avoid rebuilding of the gradient tree.

    Arguments:
        func (Function): Function to memoize.
    """
    cache = {}

    @wraps(func)
    def memoizer(op_tree, be, next_error=None):
        """
        If params in the caches, return results directly. Othewise, add to cache
        and return the results.

        Arguments:
            op_tree (OpTreeNode): the op-tree to supply to the func.
            be (Backend): computation backend to supply to the func.
            next_error (Tensor or OpTreeNode, optional): next layer's error to
                                                         supply to the func.
        """
        key = (op_tree.key(), be, next_error)
        if key not in cache:
            cache[key] = func(op_tree, be, next_error)
            # print 'created grad_tree cache'
        return cache[key]
    return memoizer


@memoize_autodiff
class Autodiff(object):

    """
    Automatic differentiation given an op-tree.

    Arguments:
        op_tree (OpTreeNode): the op-tree to take gradient of
        be (Backend): computation backend used
        next_error (Tensor or OpTreeNode, optional): next layer's error, usually
                                                     self.delta in a layer. If
                                                     set to None, then automatically
                                                     the default value is tensor
                                                     ones() in output shape
    """

    __slots__ = ['op_tree', 'be', 'dtype', 'next_error', 'map_tensor_grad_node',
                 'map_tensor_grad_op_tree', 'grad_node']

    def __init__(self, op_tree, be, next_error=None):
        # check type
        assert (type(op_tree) in _scalar_types or type(op_tree) == OpTreeNode or
                isinstance(op_tree, Tensor)), "op_tree type not supported"
        assert be is not None

        # attributes
        self.op_tree = op_tree
        self.be = be
        self.dtype = be.default_dtype
        if next_error is not None:
            assert next_error.shape == op_tree.shape, (
                "next_error.shape %s must be consistant with op_tree.shape %s"
                % (next_error.shape, op_tree.shape))
            self.next_error = next_error
        else:
            self.next_error = self.be.ones(op_tree.shape)

        self.map_tensor_grad_node = {}  # for building grad tree
        self.map_tensor_grad_op_tree = {}  # quick access to grad_op_tree

        # build_grad
        self.grad_node = GradNode(op_tree, self)
        if self.next_error:
            self.grad_node.grad_op_tree = self.next_error
        else:
            self.grad_node.grad_op_tree = self.be.ones(self.op_tree.shape)
        self.grad_node.build_grad()

    def __del__(self):
        self.cleanup()

    def cleanup(self):
        """
        Perform cleanup on object deletion.
        """
        if self.grad_node is not None:
            self.grad_node.cleanup()
        self.grad_node = None
        self.dtype = None
        self.next_error = None
        self.op_tree = None
        self.be = None

    def back_prop_grad(self, tensors, gradients):
        """
        Back-propagate the gradient of the `tensors` to `gradients`.

        Arguments:
            Tensors (list): List of Tensors to compute gradients.
            Gradient (list): List of Tensors, as output buffers of the
                             Gradients.
        """
        # avoid tensor reused as a grad_buffer
        for grad_buffer in gradients:
            assert(grad_buffer._original_base not in self.map_tensor_grad_op_tree)

        skipped_tensor = None
        for tensor, grad_buffer in zip(tensors, gradients):
            if grad_buffer is self.next_error:
                # next_error reused as a grad_buffer
                skipped_tensor = tensor
            else:
                grad_buffer[:] = self.map_tensor_grad_op_tree.get(
                    tensor._original_base, grad_buffer * 0.)

        if skipped_tensor:
            self.next_error[:] = self.map_tensor_grad_op_tree.get(
                skipped_tensor._original_base, self.next_error * 0.)

    def get_grad_op_tree(self, tensors):
        """
        Get gradient op_trees w.r.t the list of `tensors`. If a tensor is not
        used, its gradient will be set to zero.

        Arguments:
            Tensors (list): List of Tensors to compute gradients.

        Returns
            list: A list of op_trees, each of them is the gradent of the input
                  tensor.
        """
        grad_op_trees = []
        for tensor in tensors:
            grad_op_trees.append(
                self.map_tensor_grad_op_tree.get(tensor._original_base, tensor * 0.))
        return grad_op_trees

    def get_grad_tensor(self, tensors):
        """
        Get gradient values in type Tensor w.r.t the list of `tensors`. If a
        tensor is not used, its gradient will be set to zero.

        Arguments:
            Tensors (list): List of Tensors to compute gradients on.

        Returns
            list: A list of Tensors, each of them is the gradent of the input
                  tensor.
        """
        grad_op_trees = self.get_grad_op_tree(tensors)
        grad_vals = []
        for grad_op_tree in grad_op_trees:
            grad_val = self.be.empty(grad_op_tree.shape)
            grad_val[:] = grad_op_tree
            grad_vals.append(grad_val)
        return grad_vals

    def get_grad_asnumpyarray(self, tensors):
        """
        Get gradient values as numpy array w.r.t the list of `tensors`. If a
        tensor is not used, its gradient will be set to zero.

        Arguments:
            Tensors (list): List of Tensors to compute gradients.

        Returns
            list: A list of numpy.ndarray, each of them is the gradient of the
                  input tensor.
        """
        grad_vals = self.get_grad_tensor(tensors)
        for i in range(len(grad_vals)):
            grad_vals[i] = grad_vals[i].get().astype(self.dtype)
        return grad_vals


class GradNode(object):

    """
    A node in grad_tree. A GradNode contains the op_optree and the grad_op_tree
    at this location of the grad_tree, and it also has pointers to the left and
    right child in the grad_tree.
    """

    __slots__ = ['op_tree', 'grad_op_tree', 'ad', 'left', 'right']

    def __init__(self, op_tree, ad):
        """
        Arguments:
            op_tree (OpTreeNode or Tensor): the op_tree at this grad_node
            ad (Autodiff): the autodiff object with global op_tree, next_error and dicts
        """
        # check op_tree
        assert op_tree is not None

        # attributes
        self.op_tree = op_tree  # forward op_tree
        self.grad_op_tree = None  # backward gradient op_tree
        self.ad = ad  # info about autodiff object
        self.left = None
        self.right = None

        # build GradNode recursively
        if isinstance(op_tree, Tensor):
            # save to ad.map_tensor_grad_node
            if op_tree._original_base not in ad.map_tensor_grad_node:
                ad.map_tensor_grad_node[op_tree._original_base] = self
        elif type(op_tree) == OpTreeNode:
            # init recursively
            if op_tree[1] is not None:
                if (isinstance(op_tree[1], Tensor) and
                        op_tree[1]._original_base in ad.map_tensor_grad_node):
                    # seen tensor before
                    self.left = ad.map_tensor_grad_node[op_tree[1]._original_base]
                else:
                    # build recursively
                    self.left = GradNode(op_tree[1], ad)
            if op_tree[2] is not None:
                if (isinstance(op_tree[2], Tensor) and
                        op_tree[2]._original_base in ad.map_tensor_grad_node):
                    # seen tensor before
                    self.right = ad.map_tensor_grad_node[op_tree[2]._original_base]
                else:
                    # build recursively
                    self.right = GradNode(op_tree[2], ad)

    def __del__(self):
        self.cleanup()

    def cleanup(self):
        """
        Perform cleanup on object deletion.
        """
        self.op_tree = None
        self.grad_op_tree = None
        self.ad = None

        if self.left is not None:
            self.left.cleanup()
        self.left = None

        if self.right is not None:
            self.right.cleanup()
        self.right = None

    def build_grad(self):
        """
        Actually back-propagate the gradient.
        """
        # self.grad_op_tree shall be set by ad or parent grad_node
        assert self.grad_op_tree is not None

        if type(self.op_tree) == OpTreeNode:
            # get increment
            (left_increment, right_increment) = GradUtil.get_grad_back(self)

            # left increment
            if self.left.grad_op_tree is None:
                self.left.grad_op_tree = left_increment
            else:
                self.left.grad_op_tree = self.left.grad_op_tree + \
                    left_increment

            # left recursive
            self.left.build_grad()

            # check if right increment
            if right_increment is None:
                return

            # right increment
            if self.right.grad_op_tree is None:
                self.right.grad_op_tree = right_increment
            else:
                self.right.grad_op_tree = self.right.grad_op_tree + \
                    right_increment

            # right recursive
            self.right.build_grad()

        elif isinstance(self.op_tree, Tensor):
            self.ad.map_tensor_grad_op_tree[self.op_tree._original_base] = self.grad_op_tree



# ----------------------------------------------------------------------------
# Copyright 2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import sys
from neon import logger as neon_logger
from neon.data.dataloaderadapter import DataLoaderAdapter


try:
    from aeon import DataLoader as AeonLoader
except ImportError:
    neon_logger.error('Unable to load Aeon data loading module.')
    neon_logger.error('Please follow installation instructions at:')
    neon_logger.error('https://github.com/NervanaSystems/aeon')
    sys.exit(1)


def AeonDataLoader(config, adapter=True):
    if adapter:
        return DataLoaderAdapter(AeonLoader(config))
    else:
        return AeonLoader(config)



# ----------------------------------------------------------------------------
# Copyright 2014-2017 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

"""
Library for getting information of model, hardware, enviroment
during neon tests and benchmarking
"""


from neon import logger as neon_logger


def add_param_to_output(output_string, param_name, param):
    try:
        output_string += '{:>25}: {}\n'.format(param_name, param)
    except Exception:
        output_string += '{:>25}: UNKNOWN\n'.format(param_name)
    return output_string


def display_model_params(neon_args, neon_root_yaml):
    """
    Display model parameters
    :param      neon_args: contains command line arguments,
    :param neon_root_yaml: contains YAML elements
    """
    output_string = '\n-- INFORMATION: HYPER PARAMETERS ------\n'
    try:
        output_string = add_param_to_output(output_string,
                                            'backend',
                                            neon_args.backend)
        output_string = add_param_to_output(output_string,
                                            'batch size',
                                            neon_args.batch_size)
        output_string = add_param_to_output(output_string,
                                            'epochs',
                                            neon_args.epochs)
        output_string = add_param_to_output(output_string,
                                            'optimizer type',
                                            neon_root_yaml['optimizer']['type'])
        output_string = add_param_to_output(output_string,
                                            'learning rate',
                                            neon_root_yaml['optimizer']['config']['learning_rate'])
        output_string = add_param_to_output(output_string,
                                            'momentum coef',
                                            neon_root_yaml['optimizer']['config']['momentum_coef'])
    except Exception:
        output_string += 'Some parameters cannot be displayed\n'
    output_string += '----------------------------------------'
    neon_logger.display(output_string)


def display_cpu_information():
    """
    Display CPU information.
    Assumes all CPUs are the same.
    """
    import cpuinfo
    output_string = '\n-- INFORMATION: CPU -------------------\n'
    cpu_info = cpuinfo.get_cpu_info()
    try:
        output_string = add_param_to_output(output_string,
                                            'brand',
                                            cpu_info['brand'])
        output_string = add_param_to_output(output_string,
                                            'vendor id',
                                            cpu_info['vendor_id'])
        output_string = add_param_to_output(output_string,
                                            'model',
                                            cpu_info['model'])
        output_string = add_param_to_output(output_string,
                                            'family',
                                            cpu_info['family'])
        output_string = add_param_to_output(output_string,
                                            'bits',
                                            cpu_info['bits'])
        output_string = add_param_to_output(output_string,
                                            'architecture',
                                            cpu_info['arch'])
        output_string = add_param_to_output(output_string,
                                            'cores',
                                            cpu_info['count'])
        output_string = add_param_to_output(output_string,
                                            'advertised Hz',
                                            cpu_info['hz_advertised'])
        output_string = add_param_to_output(output_string,
                                            'actual Hz',
                                            cpu_info['hz_actual'])
        output_string = add_param_to_output(output_string,
                                            'l2 cache size',
                                            cpu_info['l2_cache_size'])
    except Exception:
        output_string += 'Some CPU information cannot be displayed\n'
    output_string += '----------------------------------------'
    neon_logger.display(output_string)


def display_platform_information():
    """
    Display platform information.
    """
    import platform
    output_string = '\n-- INFORMATION: PLATFORM & OS ---------\n'

    try:

        output_string = add_param_to_output(output_string,
                                            'OS',
                                            platform.platform())
        output_string = add_param_to_output(output_string,
                                            'OS release version',
                                            platform.version())
        output_string = add_param_to_output(output_string,
                                            'machine',
                                            platform.machine())
        output_string = add_param_to_output(output_string,
                                            'node',
                                            platform.node())
        output_string = add_param_to_output(output_string,
                                            'python version',
                                            platform.python_version())
        output_string = add_param_to_output(output_string,
                                            'python build',
                                            platform.python_build())
        output_string = add_param_to_output(output_string,
                                            'python compiler',
                                            platform.python_compiler())

    except Exception:
        output_string += 'Some platform information cannot be displayed\n'
    output_string += '----------------------------------------'
    neon_logger.display(output_string)



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
CPU backend layers
"""
from __future__ import division
from builtins import object
import math
from operator import mul
import numpy as np
from functools import reduce


def ceil_div(x, y):
    """
    same as int(ceil(float(x)/y)), so no need to import math lib
    """
    return -(-x // y)


class ConvLayer(object):

    """
    ConvLayer parameter object.
    This then is passed as an argument to all the convolution operations.

    N: Number of images in mini-batch
    C: Number of input feature maps
    K: Number of output feature maps

    D: Depth  of input image
    H: Height of input image
    W: Width  of input image

    T: Depth  of filter kernel
    R: Height of filter kernel
    S: Width  of filter kernel

    padding: amount of zero-padding around the given edge
    strides: factor to step the filters by in a given direction
    dilation: dilation factor for each dimension
    """

    def __init__(self, lib, dtype,
                 N, C, K,
                 D=1, H=1, W=1,
                 T=1, R=1, S=1,
                 pad_d=0, pad_h=0, pad_w=0,
                 str_d=1, str_h=1, str_w=1,
                 dil_d=1, dil_h=1, dil_w=1):

        # Compute the output spatial dimensions
        M = lib.output_dim(D, T, pad_d, str_d, dilation=dil_d)
        P = lib.output_dim(H, R, pad_h, str_h, dilation=dil_h)
        Q = lib.output_dim(W, S, pad_w, str_w, dilation=dil_w)

        self.C = C
        self.K = K
        self.M = M
        self.P = P
        self.Q = Q
        self.NCK = (N, C, K)
        self.TRS = (T, R, S)
        self.DHW = (D, H, W)
        self.MPQ = (M, P, Q)
        self.padding = (pad_d, pad_h, pad_w)
        self.strides = (str_d, str_h, str_w)
        self.dilation = (dil_d, dil_h, dil_w)

        self.dimI = (C, D, H, W, N)
        self.dimF = (C, T, R, S, K)
        self.dimO = (K, M, P, Q, N)
        self.dimS = (K, 1)
        self.dimI2 = (C * D * H * W, N)
        self.dimF2 = (C * T * R * S, K)
        self.dimO2 = (K * M * P * Q, N)
        self.sizeI = reduce(mul, self.dimI, 1)
        self.sizeF = reduce(mul, self.dimF, 1)
        self.sizeO = reduce(mul, self.dimO, 1)
        self.nOut = reduce(mul, self.MPQ, 1) * K

        if all(x == 1 for x in self.TRS) and \
           all(p == 0 for p in self.padding) and \
           all(s == 1 for s in self.strides) and \
           all(d == 1 for d in self.dilation):
            self.dot = True
        else:
            self.dot = False

            self.mSlice = [self.fprop_slice(m, T, D, pad_d, str_d, dil_d) for m in range(M)]
            self.pSlice = [self.fprop_slice(p, R, H, pad_h, str_h, dil_h) for p in range(P)]
            self.qSlice = [self.fprop_slice(q, S, W, pad_w, str_w, dil_w) for q in range(Q)]
            self.dSlice = [self.bprop_slice(d, T, M, pad_d, str_d, dil_d) for d in range(D)]
            self.hSlice = [self.bprop_slice(h, R, P, pad_h, str_h, dil_h) for h in range(H)]
            self.wSlice = [self.bprop_slice(w, S, Q, pad_w, str_w, dil_w) for w in range(W)]
        self.is_mklop = False

    def get_is_mklop(self):
        return self.is_mklop

    def set_is_mklop(self):
        self.is_mklop = True

    def set_not_mklop(self):
        self.is_mklop = False

    def fprop_slice(self, q, S, X, padding, stride, dilation):
        f1 = None
        qs = q * stride - padding
        for s in range(S):
            x = qs + s * dilation
            if f1 is None and x >= 0 and x < X:
                x1 = x
                f1 = s
            if x < X:
                x2 = x
                f2 = s
        if f1 is None:
            return (slice(0, 0, 1), slice(0, 0, 1), 0)
        return (slice(f1, f2 + 1), slice(x1, x2 + 1, dilation), f2 - f1 + 1)

    def bprop_slice(self, x, S, Q, padding, stride, dilation):
        qs = x - (dilation * (S - 1) - padding)
        f1 = None
        for s in range(S):
            q = qs + s * dilation
            if q % stride == 0:
                q //= stride
                if q >= 0 and q < Q:
                    if f1 is None:
                        f1 = s
                        x1 = q
                    f2 = s
                    x2 = q
        if f1 is None:
            return (slice(0, 0, 1), slice(0, 0, 1), 0)

        f_step = 1
        while ((f_step * dilation) % stride) != 0:
            f_step += 1
        x_step = (f_step * dilation) // stride
        return (slice(f1, f2 + 1, f_step), slice(x1, x2 + 1, x_step), 0)

    def compound_ops(self, O, X, bias, bsum, relu, brelu, slope):
        if bias is not None:
            O[:] = (O.reshape((O.shape[0], -1)) + bias).reshape(O.shape)
        if relu or brelu:
            if relu:
                if slope:
                    O[:] = np.maximum(O, 0.0) + slope * np.minimum(O, 0.0)
                else:
                    O[:] = np.maximum(O, 0.0)
            else:
                if slope:
                    O[:] = O * ((X > 0.0) + slope * (X < 0.0))
                else:
                    O[:] = O * (X > 0.0)
        if bsum is not None:
            bsum[:] = np.sum(O.reshape((O.shape[0], -1)), axis=1, keepdims=True)

    def xprop_conv(self, I, F, O, X=None, bias=None, bsum=None, alpha=1.0, beta=0.0,
                   relu=False, brelu=False, slope=0.0, backward=False, layer_op=None):

        if X is None:
            X = O

        if backward:
            I = I._tensor.reshape(self.dimO)
            O = O._tensor.reshape(self.dimI)
            X = X._tensor.reshape(self.dimI)
        else:
            I = I._tensor.reshape(self.dimI)
            O = O._tensor.reshape(self.dimO)
            X = X._tensor.reshape(self.dimO)
        F = F._tensor.reshape(self.dimF)
        if bias is not None:
            bias = bias._tensor.reshape((O.shape[0], 1))
        if bsum is not None:
            bsum = bsum._tensor.reshape((O.shape[0], 1))

        # 1x1 conv can be cast as a simple dot operation
        if self.dot:
            C = F.shape[0]
            K = F.shape[-1]
            if backward:
                # CxHWN = CxK . KxHWN
                F = F.reshape((C, K))
                I = I.reshape((K, -1))
            else:
                # KxHWN = CxK.T . CxHWN
                F = F.reshape((C, K)).T
                I = I.reshape((C, -1))

            if beta:
                O[:] = alpha * np.dot(F, I).reshape(O.shape) + beta * X
            else:
                O[:] = np.dot(F, I).reshape(O.shape)
                self.compound_ops(O, X, bias, bsum, relu, brelu, slope)
            return

        if backward:
            # C <=> K and mirror T, R, S  (0, 1, 2, 3, 4) => (4, 1, 2, 3, 0)
            F = np.transpose(F[:, ::-1, ::-1, ::-1], (4, 1, 2, 3, 0)).copy()
            mSlice, pSlice, qSlice = self.dSlice, self.hSlice, self.wSlice
        else:
            mSlice, pSlice, qSlice = self.mSlice, self.pSlice, self.qSlice

        K, M, P, Q, N = O.shape

        for m in range(M):
            sliceT, sliceD, _ = mSlice[m]
            for p in range(P):
                sliceR, sliceH, _ = pSlice[p]
                for q in range(Q):
                    sliceS, sliceW, _ = qSlice[q]

                    slicedF = F[:, sliceT, sliceR, sliceS].reshape((-1, K))
                    slicedI = I[:, sliceD, sliceH, sliceW].reshape((-1, N))

                    if beta:
                        O[:, m, p, q] = alpha * np.dot(slicedF.T, slicedI) + \
                            beta * X[:, m, p, q]
                    else:
                        O[:, m, p, q] = np.dot(slicedF.T, slicedI)

        if not beta:
            self.compound_ops(O, X, bias, bsum, relu, brelu, slope)

    # grad_bias is added for convolution layer with bias
    def update_conv(self, I, E, U, alpha=1.0, beta=0.0, grad_bias=None, layer_op=None):

        C = self.C
        K, M, P, Q, N = self.dimO

        I = I._tensor.reshape(self.dimI)
        E = E._tensor.reshape(self.dimO)
        U = U._tensor.reshape(self.dimF)

        if grad_bias is not None:
            grad = E.reshape((K, -1))
            grad_bias[:] = np.sum(grad, axis=1)

        # 1x1 conv can be cast as a simple dot operation
        if self.dot:
            # CxK = CxHWN . KxHWN.T
            I = I.reshape((C, -1))
            E = E.reshape((K, -1)).T
            if beta:
                U[:] = alpha * np.dot(I, E).reshape(U.shape) + beta * U
            else:
                U[:] = alpha * np.dot(I, E).reshape(U.shape)
            return

        if beta:
            U *= beta
        else:
            U.fill(0.0)

        for m in range(M):
            sliceT, sliceD, tlen = self.mSlice[m]
            for p in range(P):
                sliceR, sliceH, rlen = self.pSlice[p]
                for q in range(Q):
                    sliceS, sliceW, slen = self.qSlice[q]

                    slicedI = I[:, sliceD, sliceH, sliceW].reshape((-1, N))
                    slicedE = E[:, m, p, q]
                    update = np.dot(slicedI, slicedE.T).reshape((C, tlen, rlen, slen, K))
                    if alpha == 1.0:
                        U[:, sliceT, sliceR, sliceS] += update
                    else:
                        U[:, sliceT, sliceR, sliceS] += alpha * update


class DeconvLayer(ConvLayer):

    """
    DeconvLayer parameter object.
    This then is passed as an argument to all the convolution operations.

    N: Number of images in mini-batch
    C: Number of output feature maps
    K: Number of input feature maps

    P: Height of input
    Q: Width of input

    D: Depth  of output image
    H: Height of output image
    W: Width  of output image

    T: Depth  of filter kernel
    R: Height of filter kernel
    S: Width  of filter kernel

    padding: amount of zero-padding around the given edge
    strides: factor to step the filters by in a given direction
    dilation: dilation factor for each dimension
    """

    def __init__(self, lib, dtype,
                 N, C, K,
                 M, P, Q,
                 T=1, R=1, S=1,
                 pad_d=0, pad_h=0, pad_w=0,
                 str_d=1, str_h=1, str_w=1,
                 dil_d=1, dil_h=1, dil_w=1):

        tt = dil_d * (T - 1) + 1
        rr = dil_h * (R - 1) + 1
        ss = dil_w * (S - 1) + 1
        # Cannot get exact, e.g. because not unique
        D = (M - 1) * str_d - 2 * pad_d + tt
        H = (P - 1) * str_h - 2 * pad_h + rr
        W = (Q - 1) * str_w - 2 * pad_w + ss

        # Add below to get H and W tracked
        self.H = H
        self.W = W

        self.C = C
        self.K = K
        self.M = M
        self.P = P
        self.Q = Q
        self.NCK = (N, C, K)
        self.TRS = (T, R, S)
        self.DHW = (D, H, W)
        self.MPQ = (M, P, Q)
        self.padding = (pad_d, pad_h, pad_w)
        self.strides = (str_d, str_h, str_w)
        self.dilation = (dil_d, dil_h, dil_w)

        # Did not change the names of dimI, dimO, etc. even though dimI is now technically the
        # dimension of the output
        self.dimI = (C, D, H, W, N)
        self.dimF = (C, T, R, S, K)
        self.dimO = (K, M, P, Q, N)
        self.dimI2 = (C * D * H * W, N)
        self.dimF2 = (C * T * R * S, K)
        self.dimO2 = (K * M * P * Q, N)
        self.sizeI = reduce(mul, self.dimI, 1)
        self.sizeF = reduce(mul, self.dimF, 1)
        self.sizeO = reduce(mul, self.dimO, 1)
        # nOut has to change because P and Q are now the inputs
        self.nOut = reduce(mul, self.DHW, 1) * C

        if all(x == 1 for x in self.TRS) and \
           all(p == 0 for p in self.padding) and \
           all(s == 1 for s in self.strides) and \
           all(d == 1 for d in self.dilation):
            self.dot = True
        else:
            self.dot = False
            self.dSlice = [self.bprop_slice(d, T, M, pad_d, str_d, dil_d) for d in range(D)]
            self.hSlice = [self.bprop_slice(h, R, P, pad_h, str_h, dil_h) for h in range(H)]
            self.wSlice = [self.bprop_slice(w, S, Q, pad_w, str_w, dil_w) for w in range(W)]
            self.mSlice = [self.fprop_slice(m, T, D, pad_d, str_d, dil_d) for m in range(M)]
            self.pSlice = [self.fprop_slice(p, R, H, pad_h, str_h, dil_h) for p in range(P)]
            self.qSlice = [self.fprop_slice(q, S, W, pad_w, str_w, dil_w) for q in range(Q)]


class PoolLayer(object):

    """
    PoolLayer parameter object.
    This then is passed as an argument to all pooling kernels.

    op: max, avg, l2 pooling
    N: Number of images in mini-batch

    C: Number of input feature maps
    D: Depth  of input image
    H: Height of input image
    W: Width  of input image

    J: Size of feature map pooling window (maxout n_pieces)
    T: Depth  of pooling window
    R: Height of pooling window
    S: Width  of pooling window

    padding: amount of zero-padding around the given image or feature map edge
    strides: factor to step the window by in a given direction (overlap allowed)

    Leave spatial dimensions at 1 to allow feature map pooling in the fc layers.

    """

    def __init__(self, lib, dtype,
                 op, N, C,
                 D=1, H=1, W=1,
                 J=1, T=1, R=1, S=1,
                 pad_c=0, pad_d=0, pad_h=0, pad_w=0,
                 str_c=None, str_d=None, str_h=None, str_w=None):

        # default to non-overlapping
        if str_c is None:
            str_c = J
        if str_d is None:
            str_d = T
        if str_h is None:
            str_h = R
        if str_w is None:
            str_w = S

        if str_c < J or str_d < T or str_h < R or str_w < S:
            self.overlap = (math.ceil(float(J) / str_c) *
                            math.ceil(float(T) / str_d) *
                            math.ceil(float(R) / str_h) *
                            math.ceil(float(S) / str_w))
        else:
            self.overlap = 0.0

        # Compute the output dimensions
        K = lib.output_dim(C, J, pad_c, str_c, pooling=True)
        M = lib.output_dim(D, T, pad_d, str_d, pooling=True)
        P = lib.output_dim(H, R, pad_h, str_h, pooling=True)
        Q = lib.output_dim(W, S, pad_w, str_w, pooling=True)

        self.op = op
        self.C = C
        self.K = K
        self.M = M
        self.P = P
        self.Q = Q
        self.N = N
        self.JTRS = (J, T, R, S)
        self.DHW = (D, H, W)
        self.MPQ = (M, P, Q)
        self.padding = (pad_c, pad_d, pad_h, pad_w)
        self.strides = (str_c, str_d, str_h, str_w)

        self.dimI = (C, D, H, W, N)
        self.dimO = (K, M, P, Q, N)
        self.dimF2 = None
        self.dimI2 = (C * D * H * W, N)
        self.dimO2 = (K * M * P * Q, N)
        self.sizeI = reduce(mul, self.dimI, 1)
        self.sizeO = reduce(mul, self.dimO, 1)
        self.nOut = reduce(mul, self.MPQ, 1) * K

        self.kSlice = [self.pool_slice(k, J, C, pad_c, str_c) for k in range(K)]
        self.mSlice = [self.pool_slice(m, T, D, pad_d, str_d) for m in range(M)]
        self.pSlice = [self.pool_slice(p, R, H, pad_h, str_h) for p in range(P)]
        self.qSlice = [self.pool_slice(q, S, W, pad_w, str_w) for q in range(Q)]

    def pool_slice(self, q, S, X, padding, strides):
        qs = q * strides - padding
        firstI = None
        for s in range(S):
            x = qs + s
            if x >= 0 and x < X:
                if firstI is None:
                    firstI = x
                lastI = x
        return (slice(firstI, lastI + 1), lastI - firstI + 1)
